//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/PolyPtr.h
//! @brief     Defines and implements template class PolyPtr.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_POLYPTR_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_POLYPTR_H

#include "GUI/Model/Util/UtilXML.h"

class PolyBase {
public:
    QString piLabel() const { return m_label; }
    QString piTooltip() const { return m_tooltip; }
    QStringList menuEntries() const { return m_menu_entries; }

    virtual void setCurrentIndex(int index) = 0;
    virtual int currentIndex() const = 0;

protected:
    QString m_label;            //!< A label text (short, no trailing colon)
    QString m_tooltip;          //!< Tooltip text
    QStringList m_menu_entries; //!< List of options, usually presented as combo entries
};

//! Holds a polymorphous item. Possible types of the item are specified by a Catalog.

template <typename BaseItem, typename Catalog> class PolyPtr : public PolyBase {
public:
    void simpleInit(const QString& label, const QString& tooltip,
                    typename Catalog::Type currentType);

    BaseItem* certainItem() const { return m_item.get(); }

    void setCertainItem(BaseItem* t) { m_item.reset(t); }

    template <typename S> S* createCertainItem();

    void writeTo(QXmlStreamWriter* w) const;
    template <typename... Args> void readFrom(QXmlStreamReader* r, Args... args);

    void setCurrentIndex(int index) override { m_item.reset(Catalog::create(m_types[index])); }
    int currentIndex() const override { return m_types.indexOf(Catalog::type(m_item.get())); }

private:
    std::unique_ptr<BaseItem> m_item; //!< Current selection

    QVector<typename Catalog::Type> m_types = Catalog::types();
};


//! Initialize by means of a catalog class and optional creation arguments.
//!
//! The current selection will be initialized with the first type in the catalog types. The optional
//! arguments are the arguments which may be necessary for the creation method in the catalog.
template <typename B, typename C>
void PolyPtr<B, C>::simpleInit(const QString& label, const QString& tooltip,
                               typename C::Type currentType)
{
    m_label = label;
    m_tooltip = tooltip;

    m_menu_entries.clear();
    for (const auto type : m_types)
        m_menu_entries << C::uiInfo(type).menuEntry;

    int index = C::types().indexOf(currentType);
    setCurrentIndex(index);
}

//! Directly set the new item. Templated on sub item S, which may be a child of base item B.
template <typename B, typename C> template <typename S> S* PolyPtr<B, C>::createCertainItem()
{
    auto* t = new S;
    m_item.reset(t);
    return t;
}

//! Serializes the catalog index of the currently selected type and calls
//! main serialization method of the selected class.
template <typename B, typename C> void PolyPtr<B, C>::writeTo(QXmlStreamWriter* w) const
{
    const B* t = m_item.get();
    const uint typeIndex = static_cast<uint>(C::type(t));
    XML::writeAttribute(w, XML::Attrib::type, typeIndex);
    // The next line allows to see the name of item type in XML. May be skipped while reading.
    XML::writeAttribute(w, XML::Attrib::kind, C::uiInfo(C::type(t)).menuEntry);
    if (t)
        t->writeTo(w);
}

//! Deserializes the catalog index of the currently selected type, creates a new
//! object of this type and calls main deserialization method of the selected class.
template <typename B, typename C>
template <typename... Args>
void PolyPtr<B, C>::readFrom(QXmlStreamReader* r, Args... args)
{
    const uint typeIndex = XML::readUInt(r, XML::Attrib::type);
    const QString kind = XML::readString(r, XML::Attrib::kind);
    const auto type = static_cast<typename C::Type>(typeIndex);
    B* t = C::create(type, args...);
    if (t)
        t->readFrom(r);
    m_item.reset(t);
}

#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_POLYPTR_H
